open Base
open Alcotest
open Okauth_crypto
open Okweb
open Lwt.Infix

let run_handler handler req =
  Lwt_main.run (handler req >>=
                fun resp -> Response.get_body resp |> Body.read_up_to >>=
                fun body -> Lwt.return (resp, body))

let (aepk, aesk) = generate_keypair ~purpose:Encryption
let (aspk, assk) = generate_keypair ~purpose:Signing
let base_url = "http://testhost"
let config =
  let open Okauth_client in
  {
    base_url;
    app_name = "Test App";
    public_encryption_key = aepk;
    secret_encryption_key = aesk;
    public_signing_key = aspk;
    secret_signing_key = assk;
    storage = Okauth_client.In_memory_storage.empty;
  }

let test_authentication_flow =
  test_case "authentication flow" `Quick begin fun () ->
    let user_id = "testuser" in
    let okauth_host = "okauth.test" in
    let full_user_id = user_id ^ "@" ^ okauth_host in
    let (uepk, uesk) = generate_keypair ~purpose:Encryption in
    let module C = struct
      let send req =
        match Request.get_http_method req, Request.get_path req with
        | ("GET", ["v1"; uid; "public_key"]) when String.equal uid user_id ->
          let pk = Public_key.to_base64 uepk in
          Lwt.return Response.(ok |> with_string_body pk)
        | _ ->
          Lwt.return (Response.of_status 404)
    end in

    let module Client = Okauth_client.Make(C)(Okauth_client.In_memory_storage) in
    let protected_text = "it works!" in
    let ok_handler _ = Lwt.return Response.(ok |> with_string_body protected_text) in
    let test_auth_handler = Client.middleware config ok_handler in

    begin
      let (resp, _) =
        Request.get "/some-endpoint"
        |> run_handler test_auth_handler in
      (check int)
        "checking unauthenticated request status code"
        303
        (Response.get_status resp);
      (check (option string))
        "checking unauthenticated request Location"
        (Some (base_url ^ "/login"))
        (Response.find_header ~key:"Location" resp);
    end;

    begin
      let (resp, _) =
        Request.post "/login"
        |> run_handler test_auth_handler in
      (check int)
        "checking status code of POST /login without user_id"
        400
        (Response.get_status resp);
    end;

    begin
      let (resp, _) =
        Request.post "/login"
        |> Request.with_string_body "user_id=invalid_uid"
        |> run_handler test_auth_handler in
      (check int)
        "checking status code of POST /login with invalid user_id"
        400
        (Response.get_status resp);
    end;

    begin
      (check int)
        "checking that storage is empty"
        0
        Okauth_client.(In_memory_storage.size config.storage);

      let (resp, _) =
        Request.post "/login"
        |> Request.with_string_body ("user_id=" ^ full_user_id)
        |> run_handler test_auth_handler in
      let location_opt = Response.find_header resp ~key:"Location" in
      (check int)
        "checking status code of POST /login "
        303
        (Response.get_status resp);

      let okauth_req = Request.get (Option.value_exn location_opt) in
      (check (option string))
        "checking scheme of Okauth URL"
        (Some "https")
        (okauth_req |> Request.get_uri |> Uri.scheme);
      (check (option string))
        "checking host of Okauth URL"
        (Some okauth_host)
        (okauth_req |> Request.get_uri |> Uri.host);
      (check (list string))
        "checking path of Okauth URL"
        ["v1"; "testuser"; "authenticate"]
        (okauth_req |> Request.get_path);
      (check (option string))
        "checking 'k' parameter of the request to Okauth"
        (Some (Public_key.to_base64 aepk))
        (Request.find_query_param okauth_req ~key:"k");

      let opt_value_exn ~msg = function
        | Some value -> value
        | None -> fail msg
      in

      let payload =
        okauth_req
        |> Request.find_query_param ~key:"p"
        |> opt_value_exn ~msg:"'p' parameter is missing"
        |> Cipher_text.of_base64_exn in
      let nonce =
        okauth_req
        |> Request.find_query_param ~key:"n"
        |> opt_value_exn ~msg:"'n' parameter is missing"
        |> Nonce.of_base64_exn in
      let signature =
        okauth_req
        |> Request.find_query_param ~key:"s"
        |> opt_value_exn ~msg:"'s' parameter is missing"
        |> Signature.of_base64_exn in

      (check bool)
        "verifying 's' parameter of the request to Okauth"
        true
        (verify_signature (Cipher_text.to_string payload ^ Nonce.to_string nonce) signature aspk);

      let confirmation_url = decrypt payload aepk uesk nonce |> Plain_text.to_string in
      (check bool)
        "checking confirmation URL"
        true
        (String.is_prefix ~prefix:(base_url ^ "/login/confirm/") confirmation_url);
      let confirmation_req = Request.get confirmation_url in
      let confirmation_key = Request.get_path confirmation_req |> List.last_exn in
      (check bool)
        "checking that confirmation exists in the storage"
        true
        Okauth_client.(In_memory_storage.find config.storage ~key:confirmation_key |> Option.is_some);

      let (resp, _) = run_handler test_auth_handler confirmation_req in
      (check int)
        "checking status code of GET /login/confirm "
        303
        (Response.get_status resp);
      (check (option string))
        "checking Location header value of GET /login/confirm"
        (Some base_url)
        (Response.find_header resp ~key:"Location");
      (check bool)
        "checking that confirmation is removed from the storage once confirmed"
        false
        Okauth_client.(In_memory_storage.find config.storage ~key:confirmation_key |> Option.is_some);
      let resp_cookies = Response.get_cookies resp in
      let session_id_cookie =
        resp_cookies
        |> Cookie_jar.find_cookie ~key:"session_id"
        |> opt_value_exn ~msg:"'session_id' cookie is missing" in
      let session_id = Cookie.get_value session_id_cookie in
      (check bool)
        "checking that session exists in the storage"
        true
        Okauth_client.(In_memory_storage.find config.storage ~key:session_id |> Option.is_some);
      (check bool)
        "checking Expires param of session_id cookie"
        true
        (session_id_cookie
         |> Cookie.get_expires
         |> Option.exists ~f:(fun t -> Core_kernel.Time.(
             between
               ~low:(add (now ()) (Span.of_day 13.0))
               ~high:(add (now ()) (Span.of_day 15.0))
               t)));
      (check (option string))
        "checking Path param of session_id cookie"
        (Some "/")
        (Cookie.get_path session_id_cookie);
      (check bool)
        "checking HttpOnly param of session_id cookie"
        true
        (Cookie.is_http_only session_id_cookie);
      (check bool)
        "checking Secure param of session_id cookie"
        true
        (Cookie.is_secure session_id_cookie);

      let user_id_cookie =
        resp_cookies
        |> Cookie_jar.find_cookie ~key:"user_id"
        |> opt_value_exn ~msg:"'user_id' cookie is missing" in
      (check string)
        "checking user_id cookie value"
        full_user_id
        (Cookie.get_value user_id_cookie);
      (check bool)
        "checking Expires param of user_id cookie"
        true
        (user_id_cookie
         |> Cookie.get_expires
         |> Option.exists ~f:(fun t -> Core_kernel.Time.(
             between
               ~low:(add (now ()) (Span.of_day 364.0))
               ~high:(add (now ()) (Span.of_day 366.0))
               t)));
      (check (option string))
        "checking Path param of user_id cookie"
        (Some "/")
        (Cookie.get_path user_id_cookie);

      let v = Option.value_exn Okauth_client.(In_memory_storage.find config.storage ~key:session_id) in
      Caml.print_endline v;

      let (resp, body) =
        Request.get "/some-endpoint"
        |> Request.with_cookies resp_cookies
        |> run_handler test_auth_handler in
      (check int)
        "checking status code of authenticated request"
        200
        (Response.get_status resp);
      (check string)
        "checking body of authenticated request"
        protected_text
        body;
    end;
  end

module Always_failing_client = struct
  let send _ = failwith "Always_failing_client is supposed to fail. Always."
end

let test_session_expiration =
  test_case "session expiration" `Quick begin fun () ->
    let module Client = Okauth_client.Make(Always_failing_client)(Okauth_client.In_memory_storage) in
    let ok_handler _ = Lwt.return Response.ok in
    let test_auth_handler = Client.middleware config ok_handler in

    let session_id = "expired_session" in
    let session_value = "((s_user_id user@host)(s_expires(1970-01-01 00:00:00.000000+00:00)))" in
    Okauth_client.(In_memory_storage.set config.storage ~key:session_id ~value:session_value);

    let (resp, _) =
      Request.get "/some-endpoint"
      |> Request.with_cookies (Cookie_jar.of_list [
          Cookie.make ("session_id", session_id);
          Cookie.make ("user_id", "user@host");])
      |> run_handler test_auth_handler in
    (check int)
      "checking status code of authenticated request with expired session"
      303
      (Response.get_status resp);
    (check (option string))
      "checking Location header value of authenticated request with expired session"
      (Some (base_url ^ "/login"))
      (Response.find_header resp ~key:"Location");
  end

let test_confirmation_expiration =
  test_case "confirmation expiration" `Quick begin fun () ->
    let module Client = Okauth_client.Make(Always_failing_client)(Okauth_client.In_memory_storage) in
    let ok_handler _ = Lwt.return Response.ok in
    let test_auth_handler = Client.middleware config ok_handler in

    let confirmation_id = "expired_confirmation" in
    let confirmation_value = "((c_user_id user@host)(c_expires(1970-01-01 00:00:00.000000+00:00)))" in
    Okauth_client.(In_memory_storage.set config.storage ~key:confirmation_id ~value:confirmation_value);

    let (resp, _) =
      Request.get ("/login/confirm/" ^ confirmation_id)
      |> run_handler test_auth_handler in
    (check int)
      "checking status code of confirmation request with expired confirmation"
      400
      (Response.get_status resp);
  end

let test_logout =
  test_case "logout" `Quick begin fun () ->
    let module Client = Okauth_client.Make(Always_failing_client)(Okauth_client.In_memory_storage) in
    let ok_handler _ = Lwt.return Response.ok in
    let test_auth_handler = Client.middleware config ok_handler in

    let (resp, _) =
      Request.get "/logout"
      |> run_handler test_auth_handler in
    (check int)
      "checking status code of logout request"
      303
      (Response.get_status resp);
    let session_cookie = Response.get_cookies resp |> Cookie_jar.find_cookie ~key:"session_id" in
    (check (option string))
      "checking value of session_id cookie"
      (Some "")
      (session_cookie |> Option.map ~f:Cookie.get_value);
    (check (option string))
      "checking Expires param of session_id cookie"
      (Some "1970-01-01_00-00-00.000000")
      (session_cookie
       |> Option.bind ~f:Cookie.get_expires
       |> Option.map ~f:Core_kernel.Time.(to_filename_string ~zone:Zone.utc));
  end

let cases = [
  test_authentication_flow;
  test_session_expiration;
  test_confirmation_expiration;
  test_logout;
]
